#include <iostream>
#include <exception>
#include <fstream>
#include <sstream>
#include <vector>
#include "ics46goody.hpp"
#include "array_queue.hpp"
#include "stopwatch.hpp"

template<class T>
class TN {
  public:
    TN ()                        : left(nullptr), right(nullptr){}
    TN (const TN<T>& tn)         : value(tn.value), left(tn.left), right(tn.right){}
    TN (T v, TN<T>* l = nullptr,
             TN<T>* r = nullptr) : value(v), left(l), right(r){}

    T      value;
    TN<T>* left;
    TN<T>* right;
};


template<class T>
int size(TN<T>* t) {
  if (t == nullptr)
    return 0;
  else
    return 1 + size(t->left) + size(t->right);
}


template<class T>
int height (TN<T>* t) {
  if (t == nullptr)
    return -1;
  else
    return 1 + std::max(height(t->left),height(t->right));
}


template<class T>
int depth (TN<T>* t, int value, int current_depth) {
  if (t == nullptr)
    return -1;
  else if (t->value == value)
    return current_depth;
  else {
    int left_depth = depth(t->left, value, current_depth+1);
    if (left_depth != -1)
      return left_depth;
    else
      return depth(t->right, value, current_depth+1);
  }
}

template<class T>
int depth (TN<T>* t, int value)
{return depth(t,value,0);}


template<class T>
void print_rotated(TN<T>* t,std::string  indent, std::ostream& outs) {
  if (t == nullptr)
    return;
  else {
    print_rotated(t->right, indent+"..", outs);
    outs << indent << t->value << std::endl;
    print_rotated(t->left, indent+"..", outs);
  }
}


template<class T>
std::ostream& operator << (std::ostream& outs, TN<T>* t) {
  print_rotated(t,"",outs);
  return outs;
}

int binary_trees (int n) {
  if (n == 0 || n == 1)
    return 1;
  else {
    int all = 0;
    for (int left_n = 0; left_n < n; left_n++)
      all += binary_trees(left_n) * binary_trees(n-left_n-1) ;
    return all;
  }
}


template<class T>
TN<T>* find_i (TN<T>* root, int to_find) {
  for (TN<T>* c=root; c!=nullptr; c = to_find<c->value ? c->left : c->right)
    if (to_find == c->value)
      return c;

  return nullptr;
}


template<class T>
TN<T>* find_r (TN<T>* t, T to_find) {
  if (t == nullptr)
    return nullptr;
  else
    if (to_find == t->value)
      return t;
    else if (to_find < t->value)
      return find_r(t->left, to_find);
    else /* to_find > t->value */
      return find_r(t->right,to_find);
}



template<class T>
TN<T>* insert_r1 (TN<T>* t, T to_insert) {
  if (t == nullptr)
    return new TN<T>(to_insert); //nullptr implicit for left/right subtrees
  else {
    if (to_insert < t->value)
      t->left = insert_r1(t->left, to_insert);
    else if (to_insert > t->value)
      t->right = insert_r1(t->right, to_insert);
    return t;
  }
}

template<class T>
void insert_r2 (TN<T>*& t, T to_insert) {
  if (t == nullptr)
    t = new TN<T>(to_insert);    //nullptr implicit for left/right subtrees
  else
    if (to_insert < t->value)
      insert_r2(t->left, to_insert);
    else if (to_insert > t->value)
      insert_r2(t->right, to_insert);
}

template<class T>
TN<T>* insert_i (TN<T>* root, T to_insert) {
  if (root == nullptr)
    return new TN<T>(to_insert);
  else {
    TN<T>* t = root;
    for (;;)
      if (to_insert == t->value)
        return root;
      else if (to_insert < t->value) {
        if (t->left == nullptr) {
          t->left = new TN<T>(to_insert);
          return root;
        }else
          t = t->left;
      }else /* to_insert > t->value */{
        if (t->right == nullptr) {
          t->right = new TN<T>(to_insert);
          return root;
        }else
          t = t->right;
      }
  }
}


template<class T>
T remove_closest(TN<T>*& t) {
  if (t->right == nullptr) {
    T to_return = t->value;
    TN<T>* to_delete = t;
    t = t->left;
    delete to_delete;
    return to_return;
  }else
    return remove_closest(t->right);
}

template<class T>
void remove (TN<T>*& t, T to_remove) {
  if (t == nullptr)
    return;
  else
    if (to_remove == t->value) {
      if (t->left == nullptr) {
        TN<T>* to_delete = t;
        t = t->right;
        delete to_delete;
      }else if (t->right == nullptr) {
        TN<T>* to_delete = t;
        t = t->left;
        delete to_delete;
      }else
        t->value = remove_closest(t->left);
    }else
      remove( (to_remove < t->value ? t->left : t->right), to_remove);
}


template<class T>
void print_preorder (TN<T>* t) {
  if (t == nullptr)
    return;
  else {
    std::cout << t->value << " ";
    print_preorder(t->left);
    print_preorder(t->right);
  }
}

template<class T>
void print_inorder (TN<T>* t) {
  if (t == nullptr)
    return;
  else {
    print_inorder(t->left);
    std::cout << t->value << " ";
    print_inorder(t->right);
  }
}

template<class T>
void print_postorder (TN<T>* t) {
  if (t == nullptr)
    return;
  else {
    print_postorder(t->left);
    print_postorder(t->right);
    std::cout << t->value << " ";
  }
}

template<class T>
void print_breadthfirst (TN<T>* t) { //Breadth-First
  ics::ArrayQueue<TN<T>*> q;
  q.enqueue(t);                           //Initialize with the root

  while (!q.empty()) {
    TN<T>* next = q.dequeue();
    std::cout << next->value << " ";
    if (next->left != nullptr)     //Only non-nullptr values added
      q.enqueue(next->left);
    if (next->right != nullptr)
      q.enqueue(next->right);
  }
}

//empirical


//
//template<class T>
//LN<T>* build_linked_list_simple(T values[], int length) {  //O(N^2)
//  if (length == 0)
//    return nullptr;
//  LN<T>* front = nullptr;
//  for (int i=0; i<length; ++i)
//    front = add_rear(front,values[i]);
//  return front;
//}
//
//
//
template<class T>
TN<T>* copy (TN<T>* t) {
  if (t == nullptr)
    return nullptr;
  else
    return new TN<T>(t->value, copy(t->left), copy(t->right));
}


template<class T>
bool equal (TN<T>* t1, TN<T>* t2) {
  if (t1 == nullptr || t2 == nullptr)
    return t1 == nullptr && t2 == nullptr;
  else
    return t1->value == t2->value &&
           equal(t1->left,t2->left) && equal(t1->right,t2->right);
}


TN<int>* prompt_tree(std::string preface, std::string message = "  Enter next element for temp") {
  TN<int>* answer = nullptr;
  for (;;) {
    int e = ics::prompt_int(preface + message + "(-1 to quit)");
    if (e == -1)
      break;
    insert_r2(answer,e);
  }
  return answer;
}





TN<int>* t;

std::string menu_prompt (std::string preface) {
  std::cout << "\n\n"+preface+"tree =\n" << t << std::endl;
  std::cout << preface+"Mutators                   Accessors                  General" << std::endl;
  std::cout << preface+"  ir1 - insert_r1             s   - size                 b  - build"     << std::endl;
  std::cout << preface+"  ir2 - insert_r2             h   - height               lf - load file" << std::endl;
  std::cout << preface+"  ii  - insert_i              d   - depth                em - empirical" << std::endl;
  std::cout << preface+"  r   - remove                #   - # binary trees       q   - quit"     << std::endl;
  std::cout << preface+"                              fi  - find iterative"   << std::endl;
  std::cout << preface+"                              fr  - find recursive"   << std::endl;
  std::cout << preface+"                              c   - copy"             << std::endl;
  std::cout << preface+"                              =   - equals"           << std::endl;
  std::cout << preface+"                              p<  - print preorder"   << std::endl;
  std::cout << preface+"                              pi  - print inorder"    << std::endl;
  std::cout << preface+"                              p>  - print postorder"  << std::endl;
  std::cout << preface+"                              pb  - print breadth"    << std::endl;

  std::string allowable[] = {"ir1","ir2","ii","r","s","h","d","f","#","fi","fr","c","=","p<","pi","p>","pb","b","lf","em","q",""};
  return ics::prompt_string("\n"+preface+"Enter tree command","",allowable);
}

void process_commands(std::string preface) {
  for (;;) try{
    std::string command = menu_prompt(preface);

    if (command == "ir1")
      t = insert_r1(t,ics::prompt_int(preface+"  Enter element to insert"));

    else if (command == "ir2")
      insert_r2(t,ics::prompt_int(preface+"  Enter element to insert"));

    else if (command == "ii") {
      int v = ics::prompt_int(preface+"  Enter element to insert");
      if (t == nullptr)
        t = new TN<int>(v);
      else
        insert_i(t,v);
    }

    else if (command == "r")
      remove(t,ics::prompt_int(preface+"  Enter element to remove"));

    else if (command == "b")
      t = prompt_tree(preface);

    else if (command == "lf") {
      t = nullptr;
      std::ifstream t_file;
      ics::safe_open(t_file,preface+"  Enter file name to read", "load.txt");
      std::string line;
      while (getline(t_file,line)) {
        std::istringstream input(line);
        int e;
        input >> e;
        t = insert_i(t,e);
      }
    }


    else if (command == "s")
      std::cout << "  size = " << size(t) << std::endl;

    else if (command == "h")
      std::cout << "  size = " << height(t) << std::endl;

    else if (command == "d"){
      int e = ics::prompt_int(preface+"  Enter element to find depth of");
      std::cout << "  depth = " << depth(t,e) << std::endl;
    }

    else if (command == "#"){
      int n = ics::prompt_int(preface+"  Enter size of tree");
      std::cout << "  # of binary trees of that size = " << binary_trees(n) << std::endl;
    }

    else if (command == "fi"){
      int e = ics::prompt_int(preface+"  Enter element to find");
      std::cout << "  find (shows subtree rooted at element) =\n" << find_i(t,e) << std::endl;
    }

    else if (command == "fr"){
      int e = ics::prompt_int(preface+"  Enter element to find");
      std::cout << "  find (shows subtree rooted at element) =\n" << find_r(t,e) << std::endl;
    }

    else if (command == "c") {
      TN<int>* cp = copy(t);
      std::cout << preface+"  copy =\n" << cp << std::endl;
    }

    else if (command == "=") {
      TN<int>* temp = prompt_tree(preface);
      std::cout << preface+ "  equals(t,temp) = " << equal(t,temp) << std::endl;
    }

    else if (command == "p<") {
      print_preorder(t);
      std::cout << std::endl;
    }

    else if (command == "pi") {
      print_inorder(t);
      std::cout << std::endl;
    }

    else if (command == "p>") {
      print_postorder(t);
      std::cout << std::endl;
    }

    else if (command == "pb") {
      print_breadthfirst(t);
      std::cout << std::endl;
    }

    else if (command == "em") {
      std::cout << "  Empirical not implemented yet" << std::endl;
    }

    else if (command == "q")
      break;

    else
      std::cout << preface+"\""+command+"\" is unknown command" << std::endl;

  } catch (const std::exception& e) {
    std::cout << e.what() << std::endl;
  }

}

int main() {
  process_commands("");
  return 0;
}
